from __future__ import annotations
import asyncio
from concurrent.futures import ThreadPoolExecutor, as_completed
from fast_depends import inject
from fast_depends.core import CallModel
from fast_depends.dependencies import model
from loguru import logger
from abc import ABC, abstractmethod
import collections
import functools
import inspect
import tqdm
import tqdm.asyncio as tqdm_asyncio
import typing
from loguru import logger

# from functools import up
from copy import deepcopy
from result import Ok, Err, Result


def is_inject_decorated(fn):
    return (
        hasattr(fn, "__closure__")
        and len((fn.__closure__ or [])) > 1
        and isinstance(fn.__closure__[1].cell_contents, CallModel)
    )


class Output:
    def __init__(self):
        self._data = {}
        self._types = {}
        self._locked = False

    def register(self, name: str, data_type: str):
        self._types[name] = data_type

    def lock(self):
        self._locked = True

    def copy(self) -> typing.Self:
        return deepcopy(self)

    def __setattr__(self, name: str, value: typing.Any):
        if name in ["_data", "_types", "_locked"]:
            super().__setattr__(name, value)
            return

        if self._locked:
            raise AttributeError("Output is locked. Cannot add new attributes")

        if name in self._types:
            expected_type = self._types[name]
            if not self._check_type(value, expected_type):
                raise TypeError(f"Value for {name} must be " f"of type {expected_type}")

        self._data[name] = value

    def __getattr__(self, name: str) -> typing.Any:
        if name == "_data":
            raise AttributeError()
        if name in self._data:
            return self._data[name]
        raise AttributeError(f"{name} not found")

    def _check_type(self, value: typing.Any, expected_type: str) -> bool:
        origin_type = typing.get_origin(expected_type)
        type_args = typing.get_args(expected_type)

        if isinstance(origin_type, type) and origin_type in [list, set, tuple]:
            if not isinstance(value, origin_type):
                return False
            return all(self._check_type(item, type_args[0]) for item in value)

        if origin_type is dict:
            if not isinstance(value, dict):
                return False
            key_type, val_type = type_args
            return all(
                self._check_type(k, key_type) and self._check_type(v, val_type)
                for k, v in value.items()
            )

        if origin_type is typing.Union:
            return any(self._check_type(value, arg) for arg in type_args)

        return type(value).__name__ == expected_type

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._data.keys())})"

    def as_dict(self) -> dict[str, typing.Any]:
        return self._data.copy()


class PipelineResult:
    def __init__(
        self,
        sync_generator: typing.Generator[Result[Output, Exception], None, None]
        | None = None,
        async_generator: typing.AsyncGenerator[Result[Output, Exception], None]
        | None = None,
    ):
        self._sync_generator = sync_generator
        self._async_generator = async_generator

    def __iter__(self) -> typing.Iterator[Result[Output, Exception]]:
        if self._sync_generator:
            return iter(self._sync_generator)
        else:
            raise TypeError("Cannot use __iter__ on an async pipeline result")

    async def __aiter__(self) -> typing.AsyncIterator[Result[Output, Exception]]:
        if self._async_generator:
            async for result in self._async_generator:
                yield result
        elif self._sync_generator:
            # Wrap the synchronous generator in an async generator
            for result in self._sync_generator:
                yield result
        else:
            raise TypeError("No generator available for iteration")

    def __repr__(self) -> str:
        return f"PipelineResult(sync_generator={self._sync_generator}, async_generator={self._async_generator})"


_ANONYMOUS = "_anonymous"

T = typing.TypeVar("T", bound=typing.Callable, covariant=True)


# @typing.runtime_checkable
class IComponent(typing.Protocol[T]):
    @property
    def __name__(self) -> str: ...

    def __call__(self, *args, **kwargs) -> typing.Any: ...


class AbstractComponent(ABC, IComponent):
    @property
    def __name__(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    def __call__(self, *args, **kwargs) -> typing.Any:
        pass


class Attribute:
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return f"Attribute({self.name})"


class Pipeline:
    def __init__(self, *fns: IComponent, debug: bool = False, **defaults):
        self._defaults = defaults
        self._pipeline = []
        self._steps = 0
        self._slot_names = []
        self._slots = {}
        self._output = Output()
        self._func_metadata = {}
        self._debug = debug

        for fn in fns:
            self.apply(fn)

    def apply(self, fn: IComponent):
        if not is_inject_decorated(fn):
            fn = inject(fn)
        func_idx = self._register_function(fn)
        assert func_idx == self._steps
        self._steps += 1

    def _register_function(self, fn: IComponent) -> int:
        if fn in self._func_metadata:
            raise ValueError(f"Function {fn.__name__} already registered")
        self._pipeline.append(fn)
        func_idx = len(self._pipeline) - 1
        _type_reference = collections.defaultdict(set)

        input_spec = self._get_function_input_spec(fn)

        for name, _type in input_spec.items():
            _type_reference[_type].add(name)
            self._add_slot(name, _type)

        output_spec = self._get_function_return_spec(fn)

        if output_spec.get(_ANONYMOUS) is not None:
            _return_type = output_spec.get(_ANONYMOUS)
            if _return_type not in _type_reference:
                raise ValueError(
                    f"Anonymous type {_return_type} not "
                    f"found in function {fn.__name__} parameters. "
                    f"Please specify a named return type"
                )
            if len(_type_reference[_return_type]) > 1:
                raise ValueError(f"Return type {_return_type} is ambiguous")
            _name = _type_reference[_return_type].pop()
            output_spec[_name] = output_spec.pop(_ANONYMOUS)

        for name, _type in output_spec.items():
            self._add_slot(str(name), _type)

        self._func_metadata[fn] = input_spec, output_spec

        return func_idx

    def _add_slot(self, name: str, _type: typing.Type | str):
        if isinstance(name, typing.ForwardRef):
            name = name.__forward_arg__

        if isinstance(_type, str):
            _data_type_str = _type
        elif isinstance(_type, typing.Type):
            _data_type_str = _type.__name__

        if name in self._slots:
            if self._slots[name] != _data_type_str:
                raise ValueError(
                    f"Slot {name} already exists with " f"type {self._slots[name]}"
                )

        self._slot_names.append(name)
        self._slots[name] = _data_type_str
        self._output.register(name, _data_type_str)
        if name in self._defaults:
            setattr(self._output, str(name), self._defaults[name])

    def _get_function_input_spec(self, fn: IComponent):
        input_spec = {}
        sig = inspect.signature(fn)

        for name, _type in sig.parameters.items():
            if _type.default is not inspect.Parameter.empty:
                if isinstance(_type.default, model.Depends):
                    continue
                continue
            if typing.get_origin(_type.annotation) is typing.Annotated:
                _data_type, _annotation = typing.get_args(_type.annotation)
            else:
                _data_type = _type.annotation
                _annotation = None
            input_spec[name] = _data_type

        return input_spec

    def _get_function_return_spec(self, fn: IComponent):
        if hasattr(fn, "__wrapped__"):
            fn = getattr(fn, "__wrapped__")

        spec = {}

        if isinstance(fn, AbstractComponent):
            annotation = typing.get_type_hints(
                fn.__class__.__call__, include_extras=True
            )["return"]
        else:
            annotation = typing.get_type_hints(fn, include_extras=True)["return"]

        if typing.get_origin(annotation) is typing.Annotated:
            _type, annotation = typing.get_args(annotation)
        else:
            _type, annotation = annotation, None
        if typing.get_origin(_type) in (typing.Tuple, tuple):
            for _arg in typing.get_args(_type):
                _type, annotation = typing.get_args(_arg)
                if not annotation:
                    if _ANONYMOUS not in spec:
                        spec[_ANONYMOUS] = _type
                    else:
                        raise ValueError(f"Cannot have multiple " f"{_ANONYMOUS} types")
                else:
                    spec[annotation] = _type
        else:
            spec[annotation if annotation is not None else _ANONYMOUS] = _type

        return {
            k if not isinstance(k, typing.ForwardRef) else k.__forward_arg__: v.__name__
            if not isinstance(v, str)
            else v
            for k, v in spec.items()
        }

    async def run(self, **inputs):
        output = self._output.copy()

        output._data = inputs

        for fn in self._pipeline:
            input_spec, output_spec = self._func_metadata[fn]
            kwargs = {}
            for name, _type in input_spec.items():
                if name in inputs:
                    kwargs[name] = deepcopy(inputs[name])
                elif hasattr(output, name):
                    kwargs[name] = deepcopy(getattr(output, name))
                else:
                    raise ValueError(
                        f"Missing input {name} for " f"function {fn.__name__}"
                    )

            if self._debug:
                logger.debug(f"Running {fn.__name__} with inputs {kwargs}")

            if inspect.iscoroutinefunction(fn):
                fn_output = await fn(**kwargs)
            else:
                fn_output = fn(**kwargs)

            if not isinstance(fn_output, tuple):
                assert len(output_spec) == 1
                fn_output = (fn_output,)

            for _output, _name in zip(fn_output, output_spec.keys()):
                setattr(output, str(_name), _output)

        output.lock()

        return output

    async def __call__(self, *args, **inputs):
        if len(args) > 0:
            for ix in range(len(args)):
                assert self._slots[self._slot_names[ix]] == type(args[ix]).__name__
                if self._slot_names[ix] in inputs:
                    raise ValueError(
                        f"Slot {self._slot_names[ix]} " "already specified in inputs"
                    )
                inputs[self._slot_names[ix]] = args[ix]

        return await self.run(**inputs)

    def copy(self) -> typing.Self:
        pl = self.__class__(**self._defaults)
        pl._pipeline = self._pipeline.copy()
        pl._steps = self._steps
        pl._slot_names = self._slot_names.copy()
        pl._slots = self._slots.copy()
        pl._output = self._output
        pl._func_metadata = self._func_metadata.copy()
        return pl

    async def async_execute(
        self,
        *inputs: dict[str, typing.Any],
        max_workers: int = 10,
        preserve_order: bool = True,
        display_progress: bool = False,
    ) -> typing.AsyncGenerator[Result[Output, Exception], None, None]:
        # This is the async execution logic
        with ThreadPoolExecutor(max_workers=max_workers) as _executor:
            futures = []
            for input in inputs:
                pl = self.copy()
                future = asyncio.ensure_future(pl(**input))
                futures.append(future)

            if preserve_order:
                for future in tqdm_asyncio.tqdm(
                    futures, disable=not display_progress, total=len(futures)
                ):
                    try:
                        r = await future
                        yield Ok(r)
                    except Exception as e:
                        yield Err(e)
            else:
                for future in tqdm_asyncio.tqdm.as_completed(futures):
                    try:
                        r = await future
                        yield Ok(r)
                    except Exception as e:
                        yield Err(e)

    def sync_execute(
        self,
        *inputs: dict[str, typing.Any],
        max_workers: int = 10,
        preserve_order: bool = True,
        display_progress: bool = False,
    ) -> typing.Generator[Result[Output, Exception], None, None]:
        # This is the sync execution logic
        with ThreadPoolExecutor(max_workers=max_workers) as _executor:
            futures = []
            for input in inputs:
                pl = self.copy()
                future = _executor.submit(pl, **input)
                futures.append(future)

            if preserve_order:
                for future in tqdm.tqdm(
                    futures, disable=not display_progress, total=len(futures)
                ):
                    try:
                        r = future.result()
                        yield Ok(r)
                    except Exception as e:
                        yield Err(e)
            else:
                for future in tqdm.tqdm(
                    as_completed(futures),
                    disable=not display_progress,
                    total=len(futures),
                ):
                    try:
                        r = future.result()
                        yield Ok(r)
                    except Exception as e:
                        yield Err(e)

    def execute(
        self,
        *inputs: dict[str, typing.Any],
        max_workers: int = 10,
        preserve_order: bool = True,
        display_progress: bool = False,
    ) -> PipelineResult:
        # Determine if we need to run async or sync
        is_async = any(asyncio.iscoroutinefunction(fn) for fn in self._pipeline)

        if is_async:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                return PipelineResult(
                    async_generator=self.async_execute(
                        *inputs,
                        max_workers=max_workers,
                        preserve_order=preserve_order,
                        display_progress=display_progress,
                    )
                )
            else:

                async def collect_results():
                    results = []
                    async for result in self.async_execute(
                        *inputs,
                        max_workers=max_workers,
                        preserve_order=preserve_order,
                        display_progress=display_progress,
                    ):
                        results.append(result)
                    return results

                results = loop.run_until_complete(collect_results())
                logger.debug(results)
                return PipelineResult(sync_generator=(result for result in results))

        else:
            return PipelineResult(
                sync_generator=self.sync_execute(
                    *inputs,
                    max_workers=max_workers,
                    preserve_order=preserve_order,
                    display_progress=display_progress,
                )
            )


def pipeline(*fns: IComponent, **defaults):
    return Pipeline(*fns, **defaults)


def function_mapper(fn: IComponent, **mapping):
    @functools.wraps(fn)
    def _wrapper(**kwargs):
        for key, value in mapping.items():
            if key in kwargs:
                kwargs[value] = kwargs.pop(key)
        return fn(**kwargs)

    return _wrapper



import base64
import typing

import farmhash
import datetime
from ._dates import parse_timestamp
from loguru import logger

T = typing.TypeVar("T", bound=type)


# def type_converter(from_: type, to_: T) -> typing.Callable[[str], typing.Optional[T]]:
#     def _converter(value) -> typing.Optional[T]:
#         if not value:
#             return value
#         try:
#             match from_, to_:
#                 case (str, datetime.datetime):
#                     return parse_timestamp(value)
#                 case (str, datetime.date):
#                     dt = parse_timestamp(value)
#                     return dt.date() if dt else None
#                 case (str, int):
#                     return int(value)
#                 case (str, float):
#                     return float(value)
#                 case (str, bool):
#                     return value.lower() == "true"
#                 case (int, str):
#                     return str(value)
#                 case (float, str):
#                     return str(value)
#                 case (bool, str):
#                     return str(value)

#         except Exception as e:
#             logger.debug(e)
#             return None
#         return value

#     return _converter


def is_int(string) -> bool:
    try:
        int(string)
        return True
    except ValueError:
        return False


def to_int(v) -> typing.Optional[int]:
    try:
        return int(v)
    except ValueError:
        return None


def signed_64_to_unsigned_128(num):
    # Python automatically handles integer promotion to larger bit widths.
    # For a negative number, add 2^64 to shift its representation to the upper 64 bits of the 128-bit space.
    return num + (1 << 64) if num < 0 else num


def int_to_base62(num):
    # Characters to be used in base-62 encoding
    characters = (
        "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"  # # noqa: E501
    )

    if num == 0:
        return characters[0]

    num = signed_64_to_unsigned_128(num)

    base62 = []
    while num:
        num, remainder = divmod(num, 62)
        base62.append(characters[remainder])

    # The final string is built in reverse
    return "".join(reversed(base62))


def to_float(v) -> typing.Optional[float]:
    try:
        return float(v)
    except ValueError:
        return None


def to_numeric(v) -> typing.Optional[typing.Union[int, float]]:
    try:
        return int(v)
    except ValueError:
        try:
            return float(v)
        except ValueError:
            return None


def compressed_farmhash(s: str) -> str:
    return int_to_base62(farmhash.hash64(s))


def farmhash_compressed(s: str) -> str:
    return int_to_base62(farmhash.hash64(s))


def farmhash_bytes(s: str) -> bytes:
    return farmhash.hash64(s).to_bytes(8, "big")


def farmhash_hex(s: str) -> str:
    return farmhash_bytes(s).hex()


def b64_decode(string: typing.Any) -> str:
    """Decode a base64 string"""

    if not isinstance(string, str):
        string = str(string)

    # check if string is base64
    try:
        if base64.b64encode(base64.b64decode(string)).decode("utf-8") == string:
            return base64.b64decode(string).decode("utf-8")
    except Exception:
        pass
    return string


def b64_encode(string: typing.Any) -> str:
    """Encode a string to base64"""

    if not isinstance(string, str):
        string = str(string)

    return base64.b64encode(string.encode("utf-8")).decode("utf-8")



import os
import signal

from loguru import logger

SIGNAL_TRANSLATION_MAP = {
    signal.SIGINT: "SIGINT",
    signal.SIGTERM: "SIGTERM",
}


def name_process(name, with_pid: bool = False):
    import os
    import setproctitle

    pid = os.getpid()
    if with_pid:
        setproctitle.setproctitle(f"{name} - {pid}")
    else:
        setproctitle.setproctitle(name)
    return name


def parse_args(args: list[str]):
    parsed = {}
    i = 0
    while i < len(args):
        if args[i].startswith("--"):
            key = args[i][2:]
            if i + 1 < len(args) and not args[i + 1].startswith("--"):
                value = args[i + 1]
                if value in ("true", "false"):
                    value = value == "true"
                elif value.isdigit():
                    value = int(value)
                parsed[key.replace("-", "_")] = value
                i += 2
            else:
                raise ValueError(f"Missing value for argument {args[i]}")
        else:
            raise ValueError(f"Unexpected value without preceding argument: {args[i]}")
    return parsed


class KeyboardInterruptHandler:
    def __init__(self, propagate_to_forked_processes=None):
        """
        Constructs a context manager that suppresses SIGINT & SIGTERM signal handlers
        for a block of code.

        The signal handlers are called on exit from the block.

        Inspired by: https://stackoverflow.com/a/21919644

        :param propagate_to_forked_processes: This parameter controls behavior of this context manager
        in forked processes.
        If True, this context manager behaves the same way in forked processes as in parent process.
        If False, signals received in forked processes are handled by the original signal handler.
        If None, signals received in forked processes are ignored (default).
        """
        self._pid = os.getpid()
        self._propagate_to_forked_processes = propagate_to_forked_processes
        self._sig = None
        self._frame = None
        self._old_signal_handler_map = {}

    def __enter__(self):
        self._old_signal_handler_map = {
            sig: signal.signal(sig, self._handler)
            for sig, _ in SIGNAL_TRANSLATION_MAP.items()
        }

    def __exit__(self, exc_type, exc_val, exc_tb):
        for sig, handler in self._old_signal_handler_map.items():
            signal.signal(sig, handler)

        if self._sig is None:
            return

        self._on_signal(self._sig, self._frame)

    def _on_signal(self, signum, frame):
        if self._sig in self._old_signal_handler_map:
            self._old_signal_handler_map[self._sig](self._sig, self._frame)

    def _handler(self, sig, frame):
        self._sig = sig
        self._frame = frame

        #
        # Protection against fork.
        #
        if os.getpid() != self._pid:
            if self._propagate_to_forked_processes is False:
                logger.debug(
                    f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, {self._pid=}, calling original handler"
                )
                self._on_signal(self._sig, self._frame)
            elif self._propagate_to_forked_processes is None:
                logger.debug(
                    f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; "
                    f"PID mismatch: {os.getpid()=}, ignoring the signal"
                )
                return
            # elif self._propagate_to_forked_processes is True:
            #   ... passthrough

        logger.debug(
            f"!!! DelayedKeyboardInterrupt._handler: {SIGNAL_TRANSLATION_MAP[sig]} received; delaying KeyboardInterrupt"
        )



import typing
from types import UnionType, GenericAlias

class TypeInfo(typing.NamedTuple):
    type: typing.Type
    is_optional: bool
    
    
def annotation_extract_primary_type(annotation: typing.Any) -> TypeInfo:
    if typing.get_origin(annotation) is UnionType:
        args = typing.get_args(annotation)
        if len(args) == 2 and args[1] == type(None):
            return TypeInfo(args[0], is_optional=True)
        else:
            return TypeInfo(annotation, is_optional=False)
    else:
        return TypeInfo(annotation, is_optional=False)



import asyncio
import nest_asyncio
import sys
from typing import AsyncIterator, TypeVar, Callable, Awaitable, overload
import signal
import functools
from loguru import logger

def run_async(coroutine):
    try:
        # Try to get the current event loop
        loop = asyncio.get_event_loop()
    except RuntimeError:
        # If there's no current event loop, create a new one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    # Check if we're in a Jupyter notebook
    if "IPython" in sys.modules:
        # If so, apply nest_asyncio to allow nested use of event loops
        nest_asyncio.apply()

    # Now we can safely run our coroutine
    return loop.run_until_complete(coroutine)


async def _async_generator_timeout(async_gen, timeout):
    try:
        while True:
            try:
                item = await asyncio.wait_for(async_gen.__anext__(), timeout)
                yield item
            except StopAsyncIteration:
                break
    except asyncio.TimeoutError:
        raise asyncio.TimeoutError("Generator didn't emit a new item within the specified timeout")



T = TypeVar("T")

@overload
def async_iterator(
    func: Callable[..., AsyncIterator[T]]
) -> Callable[..., AsyncIterator[T]]: ...

@overload
def async_iterator(
    func: None = None,
    iteration_timeout: float | None = None
) -> Callable[[Callable[..., AsyncIterator[T]]], Callable[..., AsyncIterator[T]]]: ...


def async_iterator(
    func: Callable[..., AsyncIterator[T]] | None = None,
    iteration_timeout: float | None = None,
) -> Callable[..., AsyncIterator[T]] | Callable[[Callable[..., AsyncIterator[T]]], Callable[..., AsyncIterator[T]]]:
   
    def inner(async_iter_func: Callable[..., AsyncIterator[T]]) -> Callable[..., AsyncIterator[T]]:
        @functools.wraps(async_iter_func)
        async def wrapper(*args, **kwargs) -> AsyncIterator[T]:
            stop_event = asyncio.Event()

            loop = asyncio.get_running_loop()
            loop.add_signal_handler(signal.SIGINT, stop_event.set)
            
            try:
                
                if iteration_timeout is not None:
                    async_gen = _async_generator_timeout(async_iter_func(*args, **kwargs), iteration_timeout)
                else:
                    async_gen = async_iter_func(*args, **kwargs)
                
                async for item in async_gen:
                    if stop_event.is_set():
                        break
                    yield item
                    await asyncio.sleep(0)
            except* Exception as e:
                for exc in e.exceptions:
                    logger.error(exc)
            finally:
                loop.remove_signal_handler(signal.SIGINT)
        return wrapper
    
    if func is not None:
        return inner(func)
    
    return inner
    


import pytest
import orjson
from collections import defaultdict
from typing import Any
from dataclasses import dataclass
from loguru import logger

# Import the functions and classes to be tested
from good_common.utilities._collections import (
    sort_object_keys,
    object_farmhash,
    merge_dicts,
    deduplicate_dicts,
    clean_dict,
    recursive_remove_nones,
    set_defaults,
    merge_lists,
    flatten_list,
    recursive_to_dict,
    to_dict,
    map_object,
    expand_nested_json,
    recursive_default,
    recursive_get,
    recursive_convert_lists,
    index_object,
    deindex_object,
    DotDict,
    dict_to_flatdict,
    flatdict_to_dict,
    flatten_seq,
)


def test_sort_object_keys():
    obj = {"b": 2, "a": 1}
    result = sort_object_keys(obj)
    assert result == {"a": 1, "b": 2}


# Test merge_dicts
def test_merge_dicts():
    d1 = {"a": 1, "b": {"c": 2}}
    d2 = {"b": {"d": 3}, "e": 4}
    result = merge_dicts([d1, d2])
    assert result == {"a": 1, "b": {"c": 2, "d": 3}, "e": 4}

    # Test with keep_unique_values
    d1 = {"a": 1, "b": 2}
    d2 = {"a": 3, "c": 4}
    result = merge_dicts([d1, d2], keep_unique_values=True)
    assert result == {"a": [1, 3], "b": 2, "c": 4}


# Test deduplicate_dicts
def test_deduplicate_dicts():
    dicts = [{"a": 1}, {"b": 2}, {"a": 1}, {"c": 3}]
    result = deduplicate_dicts(dicts)
    assert result == [{"a": 1}, {"b": 2}, {"c": 3}]


# Test clean_dict
def test_clean_dict():
    d = defaultdict(list, {"a": [1, 2], "b": {"c": set([3, 4])}})
    result = clean_dict(d)
    assert isinstance(result, dict)
    assert result == {"a": [1, 2], "b": {"c": [3, 4]}}


# Test recursive_remove_nones
def test_recursive_remove_nones():
    d = {"a": 1, "b": None, "c": {"d": 2, "e": None}}
    result = recursive_remove_nones(d)
    assert result == {"a": 1, "c": {"d": 2}}


# Test set_defaults
def test_set_defaults():
    base = {"a": 1}
    result = set_defaults(base, b=2, c=3)
    assert result == {"a": 1, "b": 2, "c": 3}


# Test merge_lists
def test_merge_lists():
    lists = [[1, 2], [2, 3], [3, 4]]
    result = merge_lists(lists)
    assert result == [1, 2, 3, 4]


# Test flatten_list
def test_flatten_list():
    lst = [1, [2, 3, [4, 5]], 6]
    result = flatten_list(lst)
    assert result == [1, 2, 3, 4, 5, 6]


# Test recursive_to_dict
def test_recursive_to_dict():
    class TestObj:
        def __init__(self):
            self.a = 1
            self.b = {"c": 2}

    obj = TestObj()
    result = recursive_to_dict(obj)
    assert result == {"a": 1, "b": {"c": 2}}


# Test to_dict
def test_to_dict():
    @dataclass
    class TestDataclass:
        a: int
        b: dict

    obj = TestDataclass(a=1, b={"c": 2})
    result = to_dict(obj)
    assert result == {"a": 1, "b": {"c": 2}}


# Test map_object
def test_map_object():
    obj = {"a": "1", "b": "2", "c": "3"}
    result = map_object(obj, a=int, b=float)
    assert result == {"a": 1, "b": 2.0, "c": "3"}


# Test expand_nested_json
def test_expand_nested_json():
    obj = {"a": 1, "b": '{"c": 2, "d": {"e": 3}}'}
    result = expand_nested_json(obj)
    assert result == {"a": 1, "b": {"c": 2, "d": {"e": 3}}}


# Test recursive_default
def test_recursive_default():
    obj = {"a": {"b": 1}, "c": 2}
    assert recursive_default(obj, "a", "b") == 1
    assert recursive_default(obj, "c") == 2
    assert recursive_default(obj, "e", "f") is None


# Test recursive_get
def test_recursive_get():
    obj = {"a": {"b": {"c": 1}}}
    assert recursive_get(obj, "a", "b", "c") == 1
    assert recursive_get(obj, "a", "b", "d") is None


# Test recursive_convert_lists
def test_recursive_convert_lists():
    obj = {"a": {"0": "x", "1": "y"}, "b": {"0": "z"}}
    result = recursive_convert_lists(obj)
    assert result == {"a": ["x", "y"], "b": ["z"]}


# Test index_object and deindex_object
# def test_index_deindex_object():
#     obj = {"a": {"b": [1, 2, {"c": 3}]}}
#     indexed = index_object(obj)
#     assert indexed == {"a.b[0]": 1, "a.b[1]": 2, "a.b[2].c": 3}
#     deindexed = deindex_object(indexed)
#     assert deindexed == obj


def test_index_deindex_object():
    obj = {"a": {"b": [1, 2, {"c": 3}]}}
    indexed = index_object(obj)
    assert indexed == {"a.b[0]": 1, "a.b[1]": 2, "a.b[2].c": 3}
    deindexed = deindex_object(indexed)
    assert deindexed == obj

    # Additional test with more complex structure
    complex_obj = {"x": [{"y": 1}, 2, [3, {"z": 4}]]}
    complex_indexed = index_object(complex_obj)
    assert complex_indexed == {"x[0].y": 1, "x[1]": 2, "x[2][0]": 3, "x[2][1].z": 4}
    complex_deindexed = deindex_object(complex_indexed)
    assert complex_deindexed == complex_obj


# Test DotDict
def test_dotdict():
    d = DotDict({"a": 1, "b": {"c": 2}})
    assert d.a == 1
    assert d.b.c == 2
    d.d = {"e": 3}
    assert d.d.e == 3
    assert d["d"]["e"] == 3
    d.f = [{"g": 4}, {"h": 5}]
    assert d.f[0].g == 4
    assert d.f[1].h == 5

    # Test conversion back to dict
    regular_dict = d.to_dict()
    assert isinstance(regular_dict, dict)
    assert not isinstance(regular_dict["b"], DotDict)
    assert regular_dict == {
        "a": 1,
        "b": {"c": 2},
        "d": {"e": 3},
        "f": [{"g": 4}, {"h": 5}],
    }


# Test dict_to_flatdict and flatdict_to_dict
def test_dict_flatdict_conversion():
    nested = {"a": 1, "b": {"c": 2, "d": {"e": 3}}}
    flat = dict_to_flatdict(nested)
    assert flat == {("a",): 1, ("b", "c"): 2, ("b", "d", "e"): 3}
    restored = flatdict_to_dict(flat)
    assert restored == nested


# Test flatten_seq
def test_flatten_seq():
    seq = [1, [2, 3, [4, 5]], 6]
    result = list(flatten_seq(seq))
    assert result == [1, 2, 3, 4, 5, 6]
